use actix_web::{middleware::Logger, web, App, HttpServer};
use anyhow::Result;
use chrono::prelude::*;
use clap::{Parser, Subcommand};
use envconfig::Envconfig;
use jsonwebtoken::{Algorithm, DecodingKey, EncodingKey, Header, Validation};
use std::{
    collections::{HashMap, HashSet},
    fs,
    path::PathBuf,
    sync::Arc,
    time::Duration,
};
use uuid::Uuid;

#[derive(Parser)]
#[command(author, version, about, long_about = None)]
struct Cli {
    #[arg(long, value_name = "FILE")]
    config: PathBuf,

    #[arg(long, value_name = "FILE")]
    private_key: PathBuf,

    #[arg(long, value_name = "FILE")]
    public_key: PathBuf,

    #[command(subcommand)]
    command: Commands,
}

#[derive(Subcommand)]
enum Commands {
    GenJwt { id: Uuid },
    Start,
}

const ALGORITHM: Algorithm = Algorithm::RS512;

#[actix_web::main]
async fn main() -> Result<()> {
    env_logger::init();

    let cli = Cli::parse();

    let env_config = kdash_server::config::Config::init_from_env()?;

    let config = kdash_server::read_config(&cli.config)?;

    let sk = fs::read(cli.private_key)?;
    let encoding_key = EncodingKey::from_rsa_pem(&sk)?;

    let pk = fs::read(cli.public_key)?;
    let decoding_key = DecodingKey::from_rsa_pem(&pk)?;

    match cli.command {
        Commands::GenJwt { id: sub } => {
            let exp = Utc::now() + chrono::Duration::days(10 * 365);
            let exp_ts = exp.timestamp();
            let token = jsonwebtoken::encode(
                &Header::new(ALGORITHM),
                &kdash_server::Claims { exp: exp_ts, sub },
                &encoding_key,
            )?;

            println!("{}", token);
        }
        Commands::Start => {
            let mut mqtt_options = rumqttc::v5::MqttOptions::new(
                "kdash-server",
                env_config.kdash_mqtt_server,
                env_config.kdash_mqtt_port,
            );
            mqtt_options.set_credentials(
                env_config.kdash_mqtt_username,
                env_config.kdash_mqtt_password,
            );
            mqtt_options.set_keep_alive(Duration::from_secs(10));

            let (mqtt_client, mut eventloop) = rumqttc::v5::AsyncClient::new(mqtt_options, 20);

            tokio::spawn(async move {
                loop {
                    eventloop.poll().await.unwrap();
                }
            });

            HttpServer::new(move || {
                let devices = config.devices.to_owned();
                let device_ids: HashSet<Uuid> = config.devices.keys().cloned().collect();

                let devices_api: HashMap<Uuid, kdash_protocol::Config> = devices
                    .iter()
                    .map(|(k, v)| (k.to_owned(), kdash_protocol::Config::from(v.to_owned())))
                    .collect();

                let devices_api_json: HashMap<Uuid, Vec<u8>> = devices_api
                    .iter()
                    .map(|(k, v)| Ok((k.to_owned(), serde_json::to_vec(v)?)))
                    .collect::<Result<_>>()
                    .unwrap();

                let reqwest_client = reqwest::Client::default();

                App::new()
                    .app_data(web::Data::new(kdash_server::AppState {
                        devices,
                        devices_api,
                        devices_api_json,
                        device_ids,
                        jwt_decoding_key: decoding_key.to_owned(),
                        jwt_validation: Validation::new(ALGORITHM),
                    }))
                    .app_data(web::Data::new(reqwest_client))
                    .app_data(web::Data::new(kdash_server::MqttClient {
                        client: mqtt_client.to_owned(),
                        topic: env_config.kdash_mqtt_topic.to_owned(),
                        discovery_topic: env_config.kdash_mqtt_discovery_prefix.to_owned(),
                    }))
                    .wrap(Logger::default())
                    .wrap(kdash_server::JwtAuth::<kdash_server::Claims>::new(
                        Arc::new(kdash_server::Authority {
                            decoding_key: decoding_key.to_owned(),
                            validation: Validation::new(ALGORITHM),
                        }),
                    ))
                    .service(kdash_server::handlers::get_config)
                    .service(kdash_server::handlers::post_config)
                    .service(kdash_server::handlers::get_image)
            })
            .bind(("0.0.0.0", 8080))?
            .run()
            .await?;
        }
    }

    Ok(())
}
